"""Test for System model."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import System, SystemUpdate

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("cloud_enabled", "status_led_brightness_pct", "api_v1_enabled"),
    [
        (True, None, None),
        (None, 100, None),
        (None, None, True),
        (True, 100, None),
        (True, None, True),
        (None, 100, True),
        (True, 100, True),
    ],
)
async def test_system_update(
    cloud_enabled: bool,
    status_led_brightness_pct: int,
    api_v1_enabled: bool,
    snapshot: SnapshotAssertion,
):
    """Test System update."""
    data = SystemUpdate(
        cloud_enabled=cloud_enabled,
        status_led_brightness_pct=status_led_brightness_pct,
        api_v1_enabled=api_v1_enabled,
    )
    assert snapshot == data.to_dict()


async def test_system_update_raises_when_none_set():
    """Test systemUpdate raises when all values are None."""
    update = SystemUpdate(
        cloud_enabled=None, status_led_brightness_pct=None, api_v1_enabled=None
    )

    with pytest.raises(ValueError):
        update.to_dict()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["system"]),
        ("HWE-SKT", ["system"]),
        ("HWE-WTR", ["system"]),
        ("HWE-KWH1", ["system"]),
        ("HWE-KWH3", ["system"]),
        ("SDM230-wifi", ["system"]),
        ("SDM630-wifi", ["system"]),
    ],
)
async def test_system(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test System model."""
    for fixture in fixtures:
        data = System.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data
        assert snapshot == data



"""Test for HomeWizard Energy Models."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import CombinedModels, Device, Measurement, State, System

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["device"]),
        ("HWE-SKT", ["device"]),
        ("HWE-WTR", ["device"]),
        ("HWE-KWH1", ["device"]),
        ("HWE-KWH3", ["device"]),
        ("SDM230-wifi", ["device"]),
        ("SDM630-wifi", ["device"]),
    ],
)
async def test_device(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Device model."""
    for fixture in fixtures:
        data = Device.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data

        assert snapshot == data


@pytest.mark.parametrize(
    ("model"),
    [
        "HWE-P1",
        "HWE-SKT",
        "HWE-WTR",
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_combined_model(model: str, snapshot: SnapshotAssertion):
    """Test combined model."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    measurement = Measurement.from_dict(json.loads(load_fixtures(f"{model}/data.json")))

    combined = CombinedModels(
        device=device, measurement=measurement, state=None, system=None
    )
    assert snapshot == combined


async def test_combined_remaps_legacy_wifi_ssid_to_system(snapshot: SnapshotAssertion):
    """Test CombinedModels model remaps wifi_ssid to system."""
    device = Device.from_dict(json.loads(load_fixtures("HWE-SKT/device.json")))
    measurement = Measurement.from_dict(json.loads(load_fixtures("HWE-SKT/data.json")))

    combined = CombinedModels(
        device=device, measurement=measurement, state=None, system=None
    )

    assert combined.system.wifi_ssid == measurement.wifi_ssid
    assert snapshot == combined


async def test_combined_remaps_legacy_brightness_to_system(snapshot: SnapshotAssertion):
    """Test CombinedModels model remaps wifi_ssid to system."""
    device = Device.from_dict(json.loads(load_fixtures("HWE-SKT/device.json")))
    state = State.from_dict(json.loads(load_fixtures("HWE-SKT/state_all.json")))

    combined = CombinedModels(device=device, measurement=None, state=state, system=None)

    assert combined.system.status_led_brightness_pct == state.brightness / 2.55
    assert snapshot == combined


@pytest.mark.parametrize(
    ("model", "supports_state"),
    [
        ("HWE-P1", False),
        ("HWE-SKT", True),
        ("HWE-WTR", False),
        ("HWE-KWH1", False),
        ("HWE-KWH3", False),
        ("SDM230-wifi", False),
        ("SDM630-wifi", False),
    ],
)
async def test_device_supports_state(model: str, supports_state: bool):
    """Test Device model supports_state function."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device.supports_state() == supports_state


@pytest.mark.parametrize(
    ("model", "supports_identify"),
    [
        ("HWE-P1", True),
        ("HWE-SKT", True),
        ("HWE-WTR", True),
        ("HWE-KWH1", False),
        ("HWE-KWH3", False),
        ("SDM230-wifi", False),
        ("SDM630-wifi", False),
    ],
)
async def test_device_supports_identify(model: str, supports_identify: bool):
    """Test Device model supports_identify function."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device.supports_identify() == supports_identify


@pytest.mark.parametrize(
    ("model", "supports_cloud_enable"),
    [
        ("HWE-P1", True),
        ("HWE-SKT", True),
        ("HWE-WTR", True),
        ("HWE-KWH1", True),
        ("HWE-KWH3", True),
        ("SDM230-wifi", True),
        ("SDM630-wifi", True),
    ],
)
async def test_device_supports_cloud_enable(model: str, supports_cloud_enable: bool):
    """Test Device model supports_cloud_enable function."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device.supports_cloud_enable() == supports_cloud_enable


@pytest.mark.parametrize(
    ("model"),
    [
        "HWE-P1",
        "HWE-SKT",
        "HWE-WTR",
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_device_supports_reboot(model: str):
    """Test Device model supports_reboot function (always False for v1)."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device.supports_reboot() is False


@pytest.mark.parametrize(
    ("model", "supports_led_brightness"),
    [
        ("HWE-P1", False),
        ("HWE-SKT", True),
        ("HWE-WTR", False),
        ("HWE-KWH1", False),
        ("HWE-KWH3", False),
        ("SDM230-wifi", False),
        ("SDM630-wifi", False),
    ],
)
async def test_device_supports_led_brightness(
    model: str, supports_led_brightness: bool
):
    """Test Device model supports_led_brightness function."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device.supports_led_brightness() == supports_led_brightness


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-P1",
            [
                "data_minimal",
                "data",
                "data_no_gas",
                "data_single_phase",
            ],
        ),
        ("HWE-SKT", ["data"]),
        ("HWE-WTR", ["data"]),
        ("HWE-KWH1", ["data"]),
        ("HWE-KWH3", ["data"]),
        ("SDM230-wifi", ["data"]),
        ("SDM630-wifi", ["data"]),
        (
            "exceptions",
            [
                "data_invalid_external_ean",
                "data_invalid_external_data",
            ],
        ),
    ],
)
async def test_data(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Measurement model."""
    for fixture in fixtures:
        data = Measurement.from_dict(
            json.loads(load_fixtures(f"{model}/{fixture}.json"))
        )
        assert data

        assert snapshot == data


async def test_data_ignores_invalid_tariff():
    """Test Data model ignores invalid tariff."""

    measurement = Measurement.from_dict({"active_tariff": 5432})
    assert measurement
    assert measurement.tariff is None


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-SKT",
            ["state_all", "state_power_on", "state_switch_lock", "state_brightness"],
        ),
    ],
)
async def test_state(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test State model."""
    for fixture in fixtures:
        data = State.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data

        assert snapshot == data


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["system"]),
        ("HWE-SKT", ["system"]),
        ("HWE-WTR", ["system"]),
        ("HWE-KWH1", ["system"]),
        ("HWE-KWH3", ["system"]),
        ("SDM230-wifi", ["system"]),
        ("SDM630-wifi", ["system"]),
    ],
)
async def test_system(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test System model."""
    for fixture in fixtures:
        data = System.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data

        assert snapshot == data



"""Test for HomeWizard Energy."""

import asyncio
import json
from unittest.mock import AsyncMock, patch

import aiohttp
import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy import HomeWizardEnergyV1
from homewizard_energy.errors import DisabledError, RequestError, UnsupportedError

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


async def test_request_returns_str(aresponses):
    """Test JSON response is handled correctly."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        # pylint: disable=protected-access
        _, return_value = await api._request("api")
        assert isinstance(return_value, str)

        return_value = json.loads(return_value)
        assert return_value["status"] == "ok"
        await api.close()


async def test_request_internal_session(aresponses):
    """Test session is closed when created internally."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )

    api = HomeWizardEnergyV1("example.com")

    # pylint: disable=protected-access
    assert await api._request("api")
    await api.close()


async def test_request_detects_403(aresponses):
    """Test request detects disabled API."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=403,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        with pytest.raises(DisabledError):
            # pylint: disable=protected-access
            await api._request("api")

        await api.close()


async def test_request_detects_non_200(aresponses):
    """Test detects non-ok response."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        with pytest.raises(RequestError):
            # pylint: disable=protected-access
            await api._request("api")

        await api.close()


async def test_request_detects_clienterror():
    """Test other clienterror."""
    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        with (
            patch.object(session, "request", side_effect=aiohttp.ClientError),
            pytest.raises(RequestError),
        ):
            # pylint: disable=protected-access
            await api._request("api")

        await api.close()


### CombinedModels tests ###


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@pytest.mark.parametrize(
    ("model", "device", "data", "state", "system"),
    [
        ("HWE-KWH1", "device", "data", None, "system"),
        ("HWE-KWH3", "device", "data", None, "system"),
        ("HWE-P1", "device", "data", None, "system"),
        ("HWE-SKT", "device", "data", "state_all", "system"),
        ("HWE-WTR", "device", "data", None, "system"),
        ("SDM230-wifi", "device", "data", None, "system"),
        ("SDM630-wifi", "device", "data", None, "system"),
    ],
)
async def test_combined_models_with_valid_authentication(
    model: str,
    device: str,
    data: str,
    state: str | None,
    system: str,
    snapshot: SnapshotAssertion,
    aresponses,
):
    """Test combined models request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{device}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/data",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{data}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV1("example.com") as api:
        data = await api.combined()
        assert data is not None
        assert data == snapshot


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@pytest.mark.parametrize(
    ("model", "device", "data", "state", "system"),
    [
        ("HWE-KWH1", "device", "data", None, "system"),
        ("HWE-KWH3", "device", "data", None, "system"),
        ("HWE-P1", "device", "data", None, "system"),
        ("HWE-SKT", "device", "data", "state_all", "system"),
        ("HWE-WTR", "device", "data", None, "system"),
        ("SDM230-wifi", "device", "data", None, "system"),
        ("SDM630-wifi", "device", "data", None, "system"),
    ],
)
async def test_combined_models_with_valid_authentication_caches_device(
    model: str,
    device: str,
    data: str,
    state: str | None,
    system: str,
    snapshot: SnapshotAssertion,
    aresponses,
):
    """Test combined models request is successful when valid authentication is provided."""

    # Request 1
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{device}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/data",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{data}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    # Request 2, should use cache of `/api`
    aresponses.add(
        "example.com",
        "/api/v1/data",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{data}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV1("example.com") as api:
        data = await api.combined()
        assert data is not None
        assert data == snapshot

        data_2 = await api.combined()
        assert data_2 == data


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["device"]),
        ("HWE-SKT", ["device"]),
        ("HWE-WTR", ["device"]),
        ("HWE-KWH1", ["device"]),
        ("HWE-KWH3", ["device"]),
        ("SDM230-wifi", ["device"]),
        ("SDM630-wifi", ["device"]),
    ],
)
async def test_get_device_object(
    model: str, fixtures: str, snapshot: SnapshotAssertion, aresponses
):
    """Test device object is fetched and sets detected values."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = HomeWizardEnergyV1("example.com", clientsession=session)
            device = await api.device()

            assert device
            assert device.product_type == model

            assert device == snapshot

            await api.close()


async def test_get_device_used_cached_device_object(aresponses):
    """Test device object is fetched and sets detected values."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-P1/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)
        device = await api.device()

        assert device
        assert device.product_type == "HWE-P1"

        device_2 = await api.device()
        assert device_2 == device

        await api.close()


async def test_get_device_with_clear_cache_flag(aresponses):
    """Test device object is fetched and sets detected values."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-P1/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-SKT/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)
        device = await api.device()

        assert device
        assert device.product_type == "HWE-P1"

        device_2 = await api.device(reset_cache=True)
        assert device_2 != device

        assert device_2.product_type == "HWE-SKT"

        await api.close()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-P1",
            ["data", "data_minimal", "data_no_gas", "data_single_phase"],
        ),
        ("HWE-SKT", ["data"]),
        ("HWE-WTR", ["data"]),
        ("HWE-KWH1", ["data"]),
        ("HWE-KWH3", ["data"]),
        ("SDM230-wifi", ["data"]),
        ("SDM630-wifi", ["data"]),
    ],
)
async def test_get_data_object(
    model: str, fixtures: str, snapshot: SnapshotAssertion, aresponses
):
    """Test fetches data object and device object when unknown."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/device.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        aresponses.add(
            "example.com",
            "/api/v1/data",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = HomeWizardEnergyV1("example.com", clientsession=session)

            measurement = await api.measurement()
            assert measurement is not None
            assert measurement == snapshot

            await api.close()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-P1",
            ["data", "data_minimal", "data_no_gas", "data_single_phase"],
        ),
        ("HWE-SKT", ["data"]),
        ("HWE-WTR", ["data"]),
        ("HWE-KWH1", ["data"]),
        ("HWE-KWH3", ["data"]),
        ("SDM230-wifi", ["data"]),
        ("SDM630-wifi", ["data"]),
    ],
)
async def test_get_data_object_with_known_device(
    model: str, fixtures: str, snapshot: SnapshotAssertion, aresponses
):
    """Test fetches data object."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api/v1/data",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = HomeWizardEnergyV1("example.com", clientsession=session)

            # pylint: disable=protected-access
            api._detected_api_version = "v1"

            measurement = await api.measurement()
            assert measurement is not None
            assert measurement == snapshot

            await api.close()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-SKT",
            ["state_all", "state_power_on", "state_switch_lock", "state_brightness"],
        ),
    ],
)
async def test_get_state_object(
    model: str, fixtures: str, snapshot: SnapshotAssertion, aresponses
):
    """Test fetches state object and device object when unknown."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/device.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        aresponses.add(
            "example.com",
            "/api/v1/state",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = HomeWizardEnergyV1("example.com", clientsession=session)

            state = await api.state()
            assert state is not None
            assert state == snapshot

            await api.close()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-SKT", ["state_all"]),
    ],
)
async def test_state_set(
    model: str, fixtures: str, snapshot: SnapshotAssertion, aresponses
):
    """Test state set."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api/v1/state",
            "PUT",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json; charset=utf-8"},
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = HomeWizardEnergyV1("example.com", clientsession=session)

            state = await api.state(power_on=False, switch_lock=False, brightness=255)
            assert state
            assert state == snapshot
            await api.close()


@pytest.mark.parametrize(
    "model",
    [
        "HWE-P1",
        "HWE-WTR",
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_state_not_supported_with_cached_device(aresponses, model: str):
    """Test state set uses cached device to check is unsupported."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        await api.device()

        with pytest.raises(UnsupportedError):
            await api.state()

        await api.close()


@pytest.mark.parametrize(
    "model",
    [
        "HWE-P1",
        "HWE-SKT",
        "HWE-WTR",
    ],
)
async def test_identify(model: str, snapshot: SnapshotAssertion, aresponses):
    """Test identify call."""

    aresponses.add(
        "example.com",
        "/api/v1/identify",
        "PUT",
        aresponses.Response(
            text=load_fixtures(f"{model}/identify.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        state = await api.identify()
        assert state
        assert state == snapshot

        await api.close()


@pytest.mark.parametrize(
    "model",
    [
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_identify_not_supported(model: str, aresponses):
    """Test identify call when not supported."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/identify",
        "PUT",
        aresponses.Response(
            text="404 Not Found",
            status=404,
            headers={"Content-Type": "application/txt; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        with pytest.raises(UnsupportedError):
            await api.identify()

        await api.close()


@pytest.mark.parametrize(
    "model",
    [
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_identify_not_supported_with_cached_device(aresponses, model: str):
    """Test identify set uses cached device to check is unsupported."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        await api.device()

        with pytest.raises(UnsupportedError):
            await api.identify()

        await api.close()


async def test_telegram_gets_latest_telegram(aresponses, snapshot: SnapshotAssertion):
    """Telegram returns latest telegram."""

    aresponses.add(
        "example.com",
        "/api/v1/telegram",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-P1/telegram.txt"),
            status=200,
            headers={"Content-Type": "application/txt; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        telegram = await api.telegram()
        assert telegram
        assert telegram == snapshot


async def test_get_system_object(aresponses):
    """Test fetches system object and device object when unknown."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-SKT/device.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "GET",
        aresponses.Response(
            text=json.dumps({"cloud_enabled": True}),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "GET",
        aresponses.Response(
            text=json.dumps({"cloud_enabled": False}),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        system = await api.system()
        assert system
        assert system.cloud_enabled

        system = await api.system()
        assert system
        assert not system.cloud_enabled

        await api.close()


@pytest.mark.parametrize(
    "model",
    [
        "HWE-P1",
        "HWE-SKT",
        "HWE-WTR",
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_system_set(model: str, snapshot: SnapshotAssertion, aresponses):
    """Test system set."""

    aresponses.add(
        "example.com",
        "/api/v1/system",
        "PUT",
        aresponses.Response(
            text=load_fixtures(f"{model}/system.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        system = await api.system(cloud_enabled=True)
        assert system
        assert system == snapshot

        await api.close()


async def test_system_set_unsupported_arguments():
    """Test system set when no arguments are given."""

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        with pytest.raises(UnsupportedError):
            await api.system(api_v1_enabled=True)


async def test_system_routes_brightness_to_state(aresponses):
    """Test system sets brightness to state."""

    aresponses.add(
        "example.com",
        "/api/v1/state",
        "PUT",
        aresponses.Response(
            text=load_fixtures("HWE-SKT/state_brightness.json"),
            status=200,
            headers={"Content-Type": "application/json; charset=utf-8"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        api = HomeWizardEnergyV1("example.com", clientsession=session)

        system = await api.system(status_led_brightness_pct=100)
        assert system.status_led_brightness_pct == 100
        await api.close()


# pylint: disable=protected-access
async def test_request_handles_timeout():
    """Test request raises timeout when request takes too long."""

    api = HomeWizardEnergyV1("example.com")
    api._session = AsyncMock()
    api._session.request = AsyncMock(side_effect=asyncio.TimeoutError())

    with pytest.raises(RequestError):
        # pylint: disable=protected-access
        await api._request("api/v1/data")

    assert api._session.request.call_count == 3


async def test_close_when_out_of_scope():
    """Test close called when object goes out of scope."""
    api = HomeWizardEnergyV1("example.com")
    api.close = AsyncMock()

    async with api as hwe:
        assert hwe == api

    assert api.close.call_count == 1



"""Test the helper functions."""

import asyncio
from unittest.mock import AsyncMock

import pytest
from aiohttp import ClientSession

from homewizard_energy import has_v2_api

pytestmark = [pytest.mark.asyncio]


async def test_has_v2_api_true(aresponses):
    """Test if has_v2_api returns True for a v2 device."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=401,
        ),
    )

    async with ClientSession() as session:
        result = await has_v2_api("example.com", session)
        assert result is True


async def test_has_v2_api_false(aresponses):
    """Test if has_v2_api returns False for a non-v2 device."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=404,
        ),
    )

    async with ClientSession() as session:
        result = await has_v2_api("example.com", session)
        assert result is False


async def test_has_v2_api_exception():
    """Test if has_v2_api returns False when an exception occurs."""
    session = AsyncMock()
    session.get = AsyncMock(side_effect=asyncio.TimeoutError)

    result = await has_v2_api("example.com", session)
    assert result is False


async def test_has_v2_api_own_session(aresponses):
    """Test if has_v2_api opens and closes its own session."""
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=401,
        ),
    )

    result = await has_v2_api("example.com")
    assert result is True



"""Test the base class."""

import pytest

from homewizard_energy.errors import UnsupportedError
from homewizard_energy.homewizard_energy import HomeWizardEnergy

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("function", "exception"),
    [
        ("device", NotImplementedError),
        ("measurement", NotImplementedError),
        ("telegram", NotImplementedError),
        ("system", NotImplementedError),
        ("state", UnsupportedError),
        ("identify", NotImplementedError),
        ("reboot", UnsupportedError),
        ("batteries", UnsupportedError),
    ],
)
async def test_base_class_raises_notimplementederror(
    function: str, exception: Exception
):
    """Test the base class raises NotImplementedError."""
    with pytest.raises(exception):
        async with HomeWizardEnergy("host") as api:
            await getattr(api, function)()



"""python-homewizard-energy errors."""


class HomeWizardEnergyException(Exception):
    """Base error for python-homewizard-energy."""


class RequestError(HomeWizardEnergyException):
    """Unable to fulfill request.

    Raised when host or API cannot be reached.
    """


class InvalidUserNameError(HomeWizardEnergyException):
    """Invalid username.

    Raised when username is not valid, too short or too long.
    """


class ResponseError(HomeWizardEnergyException):
    """API responded unexpected."""


class NotFoundError(HomeWizardEnergyException):
    """Request not found.

    Raised when API responds with '404'
    """


class InvalidStateError(HomeWizardEnergyException):
    """Raised when the device is not in the correct state."""


class UnsupportedError(HomeWizardEnergyException):
    """Raised when the device is not supported from this library."""


class DisabledError(HomeWizardEnergyException):
    """Raised when device API is disabled. User has to enable API in app."""


class UnauthorizedError(HomeWizardEnergyException):
    """Raised when request is not authorized."""



"""Common models for HomeWizard Energy API."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from typing import Any

from awesomeversion import AwesomeVersion
from mashumaro.config import BaseConfig
from mashumaro.exceptions import MissingField
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializationStrategy

from .const import LOGGER, MODEL_TO_ID, MODEL_TO_NAME, Model
from .utils import get_awesome_version


class AwesomeVersionSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Serialization strategy for AwesomeVersion objects."""

    def serialize(self, value: AwesomeVersion) -> str:
        """Serialize AwesomeVersion object to string."""
        return str(value)  # pragma: no cover

    def deserialize(self, value: str) -> AwesomeVersion | None:
        """Deserialize string to AwesomeVersion object."""
        version = get_awesome_version(value)
        return version


class BaseModel(DataClassORJSONMixin):
    """Base model for all HomeWizard models."""

    # pylint: disable-next=too-few-public-methods
    class Config(BaseConfig):
        """Mashumaro configuration."""

        serialize_by_alias = True
        serialization_strategy = {  # noqa: RUF012
            AwesomeVersion: AwesomeVersionSerializationStrategy()
        }
        omit_none = True


class UpdateBaseModel(BaseModel):
    """Base model for all 'update' models."""

    def __post_serialize__(self, d: dict, context: dict | None = None):
        """Post serialize hook for UpdateBaseModel object."""
        _ = context  # Unused

        if not d:
            raise ValueError("No values to update")

        return d


@dataclass(kw_only=True)
class CombinedModels:
    """All values."""

    device: Device
    measurement: Measurement
    state: State | None
    system: System | None
    batteries: Batteries | None

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        device: Device,
        measurement: Measurement,
        state: State | None,
        system: System | None,
        batteries: Batteries | None = None,
    ):
        self.device = device
        self.measurement = measurement
        self.state = state
        self.system = system
        self.batteries = batteries

        # Move things around for backwards compatibility
        ## measurement.wifi_ssid -> system.wifi_ssid
        if self.measurement is not None and self.measurement.wifi_ssid is not None:
            if self.system is None:
                self.system = System()
            self.system.wifi_ssid = self.measurement.wifi_ssid

        ## measurement.wifi_strength -> system.wifi_strength_pct
        if self.measurement is not None and self.measurement.wifi_strength is not None:
            if self.system is None:
                self.system = System()
            self.system.wifi_strength_pct = self.measurement.wifi_strength

        ## state.brightness -> system.status_led_brightness_pct
        if self.state is not None and self.state.brightness is not None:
            if self.system is None:
                self.system = System()
            self.system.status_led_brightness_pct = (self.state.brightness / 255) * 100


def get_verification_hostname(model: str, serial_number: str) -> str:
    """Helper method to convert device model and serial to identifier

    The identifier is used to verify the device in the HomeWizard Energy API via HTTPS.
    """

    if model not in MODEL_TO_ID:
        raise ValueError(f"Unsupported model: {model}")

    return f"appliance/{MODEL_TO_ID[model]}/{serial_number}"


@dataclass(kw_only=True)
class Device(BaseModel):
    """Represent Device config."""

    model_name: str | None = None
    id: str | None = None

    product_name: str = field()
    product_type: str = field()
    serial: str = field()
    api_version: AwesomeVersion = field()
    firmware_version: str = field()

    @classmethod
    def __post_deserialize__(cls, obj: Device) -> Device:
        """Post deserialize hook for Device object."""
        _ = cls  # Unused

        obj.model_name = MODEL_TO_NAME.get(obj.product_type)
        obj.id = get_verification_hostname(obj.product_type, obj.serial)
        return obj

    def supports_state(self) -> bool:
        """Return if the device supports state."""
        return self.product_type == Model.ENERGY_SOCKET

    def supports_led_brightness(self) -> bool:
        """Return if the device supports LED brightness control."""
        if self.product_type in (
            Model.BATTERY,
            Model.ENERGY_SOCKET,
        ):
            return True

        # P1 Meter supports LED brightness from API v2.0.0 onwards
        if self.product_type == Model.P1_METER:
            return self.api_version.major >= 2

        return False

    def supports_cloud_enable(self) -> bool:
        """Return if the device supports cloud enable/disable."""
        return self.product_type != Model.BATTERY

    def supports_reboot(self) -> bool:
        """Return if the device supports reboot."""
        return self.api_version.major >= 2 and self.product_type != Model.BATTERY

    def supports_identify(self) -> bool:
        """Return if the device supports identify."""
        return self.product_type not in (
            Model.ENERGY_METER_1_PHASE,
            Model.ENERGY_METER_3_PHASE,
            Model.ENERGY_METER_EASTRON_SDM230,
            Model.ENERGY_METER_EASTRON_SDM630,
        )

    def supports_telegram(self) -> bool:
        """Return if the device supports telegram."""
        return self.product_type == Model.P1_METER

    def supports_batteries(self) -> bool:
        """Return if the device supports batteries."""
        return self.product_type in (
            Model.P1_METER,
            Model.ENERGY_METER_1_PHASE,
            Model.ENERGY_METER_3_PHASE,
            Model.ENERGY_METER_EASTRON_SDM230,
            Model.ENERGY_METER_EASTRON_SDM630,
        )

    def supported_battery_modes(self) -> list[Batteries.Mode] | None:
        """Return list of supported battery modes."""
        if not self.supports_batteries():
            return None

        modes = [
            Batteries.Mode.STANDBY,
            Batteries.Mode.TO_FULL,
            Batteries.Mode.ZERO,
        ]

        if self.api_version >= AwesomeVersion("2.2.0"):
            modes += [
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ]

        return modes


@dataclass(kw_only=True)
class Measurement(BaseModel):
    """Represent Measurement."""

    # Deprecated, use 'System'
    wifi_ssid: str | None = field(
        default=None,
    )
    wifi_strength: int | None = field(
        default=None,
    )

    # Generic
    energy_import_kwh: float | None = field(
        default=None,
    )
    energy_import_t1_kwh: float | None = field(
        default=None,
    )
    energy_import_t2_kwh: float | None = field(
        default=None,
    )
    energy_import_t3_kwh: float | None = field(
        default=None,
    )
    energy_import_t4_kwh: float | None = field(
        default=None,
    )
    energy_export_kwh: float | None = field(
        default=None,
    )
    energy_export_t1_kwh: float | None = field(
        default=None,
    )
    energy_export_t2_kwh: float | None = field(
        default=None,
    )
    energy_export_t3_kwh: float | None = field(
        default=None,
    )
    energy_export_t4_kwh: float | None = field(
        default=None,
    )

    power_w: float | None = field(
        default=None,
    )
    power_l1_w: float | None = field(
        default=None,
    )
    power_l2_w: float | None = field(
        default=None,
    )
    power_l3_w: float | None = field(
        default=None,
    )

    voltage_v: float | None = field(
        default=None,
    )
    voltage_l1_v: float | None = field(
        default=None,
    )
    voltage_l2_v: float | None = field(
        default=None,
    )
    voltage_l3_v: float | None = field(
        default=None,
    )

    current_a: float | None = field(
        default=None,
    )
    current_l1_a: float | None = field(
        default=None,
    )
    current_l2_a: float | None = field(
        default=None,
    )
    current_l3_a: float | None = field(
        default=None,
    )

    apparent_power_va: float | None = field(
        default=None,
    )
    apparent_power_l1_va: float | None = field(
        default=None,
    )
    apparent_power_l2_va: float | None = field(
        default=None,
    )
    apparent_power_l3_va: float | None = field(
        default=None,
    )

    reactive_power_var: float | None = field(
        default=None,
    )
    reactive_power_l1_var: float | None = field(
        default=None,
    )
    reactive_power_l2_var: float | None = field(
        default=None,
    )
    reactive_power_l3_var: float | None = field(
        default=None,
    )

    power_factor: float | None = field(
        default=None,
    )
    power_factor_l1: float | None = field(
        default=None,
    )
    power_factor_l2: float | None = field(
        default=None,
    )
    power_factor_l3: float | None = field(
        default=None,
    )

    frequency_hz: float | None = field(
        default=None,
    )

    # P1 Meter specific
    timestamp: datetime | None = field(
        default=None,
        metadata={"deserialize": lambda x: Measurement.to_datetime(x)},
    )
    protocol_version: int | None = field(
        default=None,
    )
    meter_model: str | None = field(
        default=None,
    )
    unique_id: str | None = field(
        default=None,
        metadata={"deserialize": lambda x: Measurement.hex_to_readable(x)},
    )

    tariff: int | None = field(
        default=None,
    )

    voltage_sag_l1_count: int | None = field(
        default=None,
    )
    voltage_sag_l2_count: int | None = field(
        default=None,
    )
    voltage_sag_l3_count: int | None = field(
        default=None,
    )

    voltage_swell_l1_count: int | None = field(
        default=None,
    )
    voltage_swell_l2_count: int | None = field(
        default=None,
    )
    voltage_swell_l3_count: int | None = field(
        default=None,
    )

    any_power_fail_count: int | None = field(
        default=None,
    )
    long_power_fail_count: int | None = field(
        default=None,
    )

    average_power_15m_w: float | None = field(
        default=None,
    )
    monthly_power_peak_w: float | None = field(
        default=None,
    )
    monthly_power_peak_timestamp: datetime | None = field(
        default=None,
        metadata={"deserialize": lambda x: Measurement.to_datetime(x)},
    )

    external_devices: dict[str, ExternalDevice] | None = field(
        default=None,
        metadata={
            "alias": "external",
            "deserialize": lambda list: Measurement.array_to_external_device_list(list),
        },
    )

    # Watermeter Specific
    active_liter_lpm: float | None = field(
        default=None,
    )
    total_liter_m3: float | None = field(
        default=None,
    )

    # Battery Specific
    cycles: int | None = field(
        default=None,
    )
    state_of_charge_pct: float | None = field(
        default=None,
    )

    @staticmethod
    def array_to_external_device_list(devices: list[dict]) -> dict[str, ExternalDevice]:
        """Convert external device dict to list of ExternalDevice objects."""
        rv: dict[str, ExternalDevice] = {}

        for item in devices:
            try:
                device = ExternalDevice.from_dict(item)
            except MissingField as e:
                LOGGER.error("Error converting external device: %s", e)
                continue
            rv[f"{device.type}_{device.unique_id}"] = device

        return rv

    @staticmethod
    def to_datetime(timestamp: str | int) -> datetime:
        """Convert DSRM gas-timestamp to datetime object.

        Args:
            timestamp: Timestamp string, formatted as YYMMDDHHMMSS or YYYY-MM-DDTHH:MM:SS

        Returns:
            A datetime object.
        """
        if isinstance(timestamp, int):
            # V1 API uses int for timestamp
            return datetime.strptime(str(timestamp), "%y%m%d%H%M%S")

        return datetime.fromisoformat(timestamp)

    @staticmethod
    def hex_to_readable(value: str | None) -> str | None:
        """Convert hex string to readable string, if possible.

        Args:
            value: String to convert, e.g. '4E47475955'

        Returns:
            A string formatted or original value when failed.
        """
        try:
            return bytes.fromhex(value).decode("utf-8")
        except (TypeError, ValueError):
            return value

    @classmethod
    # pylint: disable=too-many-statements
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        _ = cls  # Unused

        if "wifi_ssid" not in d:
            # This is a v2 API response, no need to remap
            return d

        d["protocol_version"] = d.get("smr_version")
        d["tariff"] = d.get("active_tariff")
        d["energy_import_kwh"] = d.get(
            "total_power_import_kwh", d.get("total_power_import_t1_kwh")
        )
        d["energy_import_t1_kwh"] = d.get("total_power_import_t1_kwh")
        d["energy_import_t2_kwh"] = d.get("total_power_import_t2_kwh")
        d["energy_import_t3_kwh"] = d.get("total_power_import_t3_kwh")
        d["energy_import_t4_kwh"] = d.get("total_power_import_t4_kwh")
        d["energy_export_kwh"] = d.get(
            "total_power_export_kwh", d.get("total_power_export_t1_kwh")
        )
        d["energy_export_t1_kwh"] = d.get("total_power_export_t1_kwh")
        d["energy_export_t2_kwh"] = d.get("total_power_export_t2_kwh")
        d["energy_export_t3_kwh"] = d.get("total_power_export_t3_kwh")
        d["energy_export_t4_kwh"] = d.get("total_power_export_t4_kwh")
        d["power_w"] = d.get("active_power_w")
        d["power_l1_w"] = d.get("active_power_l1_w")
        d["power_l2_w"] = d.get("active_power_l2_w")
        d["power_l3_w"] = d.get("active_power_l3_w")
        d["voltage_v"] = d.get("active_voltage_v")
        d["voltage_l1_v"] = d.get("active_voltage_l1_v")
        d["voltage_l2_v"] = d.get("active_voltage_l2_v")
        d["voltage_l3_v"] = d.get("active_voltage_l3_v")
        d["current_a"] = d.get("active_current_a")
        d["current_l1_a"] = d.get("active_current_l1_a")
        d["current_l2_a"] = d.get("active_current_l2_a")
        d["current_l3_a"] = d.get("active_current_l3_a")
        d["apparent_power_va"] = d.get("active_apparent_power_va")
        d["apparent_power_l1_va"] = d.get("active_apparent_power_l1_va")
        d["apparent_power_l2_va"] = d.get("active_apparent_power_l2_va")
        d["apparent_power_l3_va"] = d.get("active_apparent_power_l3_va")
        d["reactive_power_var"] = d.get("active_reactive_power_var")
        d["reactive_power_l1_var"] = d.get("active_reactive_power_l1_var")
        d["reactive_power_l2_var"] = d.get("active_reactive_power_l2_var")
        d["reactive_power_l3_var"] = d.get("active_reactive_power_l3_var")
        d["power_factor"] = d.get("active_power_factor")
        d["power_factor_l1"] = d.get("active_power_factor_l1")
        d["power_factor_l2"] = d.get("active_power_factor_l2")
        d["power_factor_l3"] = d.get("active_power_factor_l3")
        d["frequency_hz"] = d.get("active_frequency_hz")
        d["average_power_15m_w"] = d.get("active_power_average_w")
        d["monthly_power_peak_w"] = d.get("montly_power_peak_w")
        d["monthly_power_peak_timestamp"] = d.get("montly_power_peak_timestamp")
        d["external_devices"] = d.get("external_devices")

        return d

    @classmethod
    def __post_deserialize__(cls, obj: Measurement) -> Measurement:
        """Post deserialize hook for Measurement object."""
        _ = cls  # Unused

        # Some smart meters report a tariff other than 1, 2, 3 or 4, which is invalid
        if obj.tariff not in (1, 2, 3, 4):
            obj.tariff = None

        return obj


@dataclass(kw_only=True)
class ExternalDevice(BaseModel):
    """Represents externally connected device."""

    class DeviceType(StrEnum):
        """Device type allocations."""

        GAS_METER = "gas_meter"
        HEAT_METER = "heat_meter"
        WARM_WATER_METER = "warm_water_meter"
        WATER_METER = "water_meter"
        INLET_HEAT_METER = "inlet_heat_meter"

    unique_id: datetime = field(
        metadata={"deserialize": lambda x: Measurement.hex_to_readable(x)}
    )

    type: DeviceType | None = field(
        default=None,
        metadata={
            "deserialize": lambda x: ExternalDevice.DeviceType.__members__.get(
                x.upper(), None
            )
        },
    )
    value: float = field()
    unit: str = field()
    timestamp: datetime = field(
        metadata={"deserialize": lambda x: Measurement.to_datetime(x)}
    )


@dataclass(kw_only=True)
class StateUpdate(UpdateBaseModel):
    """Represent State update config."""

    power_on: bool | None = field(default=None)
    switch_lock: bool | None = field(default=None)
    brightness: int | None = field(default=None)


@dataclass(kw_only=True)
class State(BaseModel):
    """Represent current state."""

    power_on: bool | None = field(
        default=None,
    )
    switch_lock: bool | None = field(
        default=None,
    )
    brightness: int | None = field(
        default=None,
    )


@dataclass
class SystemUpdate(UpdateBaseModel):
    """Represent System update config."""

    cloud_enabled: bool | None = field(default=None)
    status_led_brightness_pct: int | None = field(default=None)
    api_v1_enabled: bool | None = field(default=None)


@dataclass(kw_only=True)
class System(BaseModel):
    """Represent System config."""

    wifi_strength_pct: int | None = None

    wifi_ssid: str | None = field(default=None)
    wifi_rssi_db: int | None = field(default=None)
    cloud_enabled: bool | None = field(default=None)
    uptime_s: int | None = field(default=None)
    status_led_brightness_pct: int | None = field(default=None)
    api_v1_enabled: bool | None = field(default=None)

    @classmethod
    def __post_deserialize__(cls, obj: System) -> System:
        _ = cls  # Unused

        if obj.wifi_rssi_db is not None:
            obj.wifi_strength_pct = (
                0
                if obj.wifi_rssi_db <= -100 or obj.wifi_rssi_db == 0
                else 100
                if obj.wifi_rssi_db >= -50
                else 2 * (obj.wifi_rssi_db + 100)
            )

        return obj


@dataclass
class BatteriesUpdate(UpdateBaseModel):
    """Represent Batteries update config."""

    mode: Batteries.Mode | None = field(default=None)
    permissions: list[Batteries.Permissions] | None = field(default=None)

    @staticmethod
    def from_mode(
        mode: Batteries.Mode,
    ) -> BatteriesUpdate:
        """Convert userfacing mode to API mode.

        Args:
            mode: Userfacing mode to convert.
        Returns:
            BatteriesUpdate object with API mode and permissions.
        """

        match mode:
            case Batteries.Mode.ZERO:
                return BatteriesUpdate(
                    mode=Batteries.Mode.ZERO,
                    permissions=[
                        Batteries.Permissions.CHARGE_ALLOWED,
                        Batteries.Permissions.DISCHARGE_ALLOWED,
                    ],
                )
            case Batteries.Mode.ZERO_CHARGE_ONLY:
                return BatteriesUpdate(
                    mode=Batteries.Mode.ZERO,
                    permissions=[
                        Batteries.Permissions.CHARGE_ALLOWED,
                    ],
                )
            case Batteries.Mode.ZERO_DISCHARGE_ONLY:
                return BatteriesUpdate(
                    mode=Batteries.Mode.ZERO,
                    permissions=[
                        Batteries.Permissions.DISCHARGE_ALLOWED,
                    ],
                )
            case Batteries.Mode.TO_FULL:
                return BatteriesUpdate(mode=Batteries.Mode.TO_FULL)
            case Batteries.Mode.STANDBY:
                return BatteriesUpdate(mode=Batteries.Mode.STANDBY, permissions=[])

        raise ValueError(f"Unsupported Batteries.Mode: {mode!r}")


@dataclass(kw_only=True)
class Batteries(BaseModel):
    """Represent Batteries config."""

    class Mode(StrEnum):
        """Device type allocations.

        This is the list of modes shown to the user. 'zero_charge_only' and 'zero_discharge_only' are
        mapped to 'zero' in the API via the Permissions parameter
        """

        ZERO = "zero"
        TO_FULL = "to_full"
        STANDBY = "standby"
        ZERO_CHARGE_ONLY = "zero_charge_only"
        ZERO_DISCHARGE_ONLY = "zero_discharge_only"

    class Permissions(StrEnum):
        """Device permission allocations."""

        CHARGE_ALLOWED = "charge_allowed"
        DISCHARGE_ALLOWED = "discharge_allowed"

    mode: Mode = field(
        metadata={
            "deserialize": lambda x: Batteries.Mode.__members__.get(x.upper(), None)
        },
    )
    permissions: list[Permissions] | None = field(
        default=None,
        metadata={
            "deserialize": lambda lst: [
                perm
                for item in lst
                if (perm := Batteries.Permissions.__members__.get(item.upper(), None))
                is not None
            ]
            if lst is not None
            else None
        },
    )
    power_w: float = field()
    target_power_w: float = field()
    max_consumption_w: float = field()
    max_production_w: float = field()
    battery_count: int | None = field(default=None)

    @classmethod
    def __post_deserialize__(cls, obj: Batteries) -> Batteries:
        """Set correct mode based on permissions after deserialization.
        If permissions is missing and mode is ZERO, keep mode as ZERO for backwards compatibility.
        If permissions is present (even if empty), apply the mapping logic.
        """
        # Only adjust if mode is ZERO
        if obj.mode != cls.Mode.ZERO:
            return obj

        # Detect if 'permissions' was present in the original data
        # If using mashumaro, the original dict is not available here, so we infer:
        # If permissions is None, treat as 'not provided'. If it's an empty list, treat as 'provided but empty'.
        if obj.permissions is None:
            # Permissions not provided, keep mode as ZERO (backwards compatibility)
            return obj

        perms = set(obj.permissions)
        if perms == {cls.Permissions.CHARGE_ALLOWED}:
            obj.mode = cls.Mode.ZERO_CHARGE_ONLY
        elif perms == {cls.Permissions.DISCHARGE_ALLOWED}:
            obj.mode = cls.Mode.ZERO_DISCHARGE_ONLY
        elif perms == {
            cls.Permissions.CHARGE_ALLOWED,
            cls.Permissions.DISCHARGE_ALLOWED,
        }:
            obj.mode = cls.Mode.ZERO
        elif perms == set():
            obj.mode = cls.Mode.STANDBY

        return obj


@dataclass(kw_only=True)
class Token(BaseModel):
    """Represent Token."""

    token: str = field()



"""HomeWizard Energy API library."""

from aiohttp import ClientSession

from .errors import DisabledError, InvalidStateError, RequestError, UnsupportedError
from .homewizard_energy import HomeWizardEnergy
from .v1 import HomeWizardEnergyV1
from .v2 import HomeWizardEnergyV2

__all__ = [
    "DisabledError",
    "HomeWizardEnergy",
    "HomeWizardEnergyV1",
    "HomeWizardEnergyV2",
    "InvalidStateError",
    "RequestError",
    "UnsupportedError",
]


async def has_v2_api(host: str, websession: ClientSession | None = None) -> bool:
    """Check if the device has support for the v2 api."""
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        # v2 api is https only and returns a 401 Unauthorized when no key provided,
        # no connection can be made if the device is not v2
        url = f"https://{host}/api"
        res = await websession.get(url, ssl=False, raise_for_status=False, timeout=5)
        res.close()

        return res.status == 401
    except Exception:  # pylint: disable=broad-except
        # all other status/exceptions means the device is not v2 or not reachable at this time
        return False
    finally:
        if not websession_provided:
            await websession.close()



"""Constants for HomeWizard Energy."""

import logging
from enum import StrEnum

LOGGER = logging.getLogger(__name__)


class Model(StrEnum):
    """Model of the HomeWizard Energy device."""

    P1_METER = "HWE-P1"
    ENERGY_SOCKET = "HWE-SKT"
    WATER_METER = "HWE-WTR"
    DISPLAY = "HWE-DSP"
    ENERGY_METER_1_PHASE = "HWE-KWH1"
    ENERGY_METER_3_PHASE = "HWE-KWH3"
    ENERGY_METER_EASTRON_SDM230 = "SDM230-wifi"
    ENERGY_METER_EASTRON_SDM630 = "SDM630-wifi"
    BATTERY = "HWE-BAT"


MODEL_TO_ID = {
    Model.P1_METER: "p1dongle",
    Model.ENERGY_SOCKET: "energysocket",
    Model.WATER_METER: "watermeter",
    Model.DISPLAY: "display",
    Model.ENERGY_METER_1_PHASE: "energymeter",
    Model.ENERGY_METER_3_PHASE: "energymeter",
    Model.ENERGY_METER_EASTRON_SDM230: "energymeter",
    Model.ENERGY_METER_EASTRON_SDM630: "energymeter",
    Model.BATTERY: "battery",
}

MODEL_TO_NAME = {
    Model.P1_METER: "Wi-Fi P1 Meter",
    Model.ENERGY_SOCKET: "Wi-Fi Energy Socket",
    Model.WATER_METER: "Wi-Fi Watermeter",
    Model.DISPLAY: "Energy Display",
    Model.ENERGY_METER_1_PHASE: "Wi-Fi kWh Meter 1-phase",
    Model.ENERGY_METER_3_PHASE: "Wi-Fi kWh Meter 3-phase",
    Model.ENERGY_METER_EASTRON_SDM230: "Wi-Fi kWh Meter 1-phase",
    Model.ENERGY_METER_EASTRON_SDM630: "Wi-Fi kWh Meter 3-phase",
    Model.BATTERY: "Plug-In Battery",
}



"""Test for Batteries model and BatteriesUpdate."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import Batteries, BatteriesUpdate

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["batteries_2_1_0", "batteries_2_2_0"]),
        ("HWE-KWH1", ["batteries"]),
        ("HWE-KWH3", ["batteries"]),
    ],
)
async def test_batteries(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Batteries model."""
    for fixture in fixtures:
        data = Batteries.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data
        assert snapshot == data


def test_batteries_battery_count_optional():
    """Test Batteries model with and without battery_count."""
    b1 = Batteries(
        mode=Batteries.Mode.ZERO,
        power_w=100.0,
        target_power_w=200.0,
        max_consumption_w=300.0,
        max_production_w=400.0,
        battery_count=None,
    )
    b2 = Batteries(
        mode=Batteries.Mode.TO_FULL,
        power_w=110.0,
        target_power_w=210.0,
        max_consumption_w=310.0,
        max_production_w=410.0,
        battery_count=2,
    )
    assert b1.battery_count is None
    assert b2.battery_count == 2


@pytest.mark.parametrize(
    "mode, expected_mode, expected_permissions",
    [
        (
            Batteries.Mode.ZERO,
            Batteries.Mode.ZERO,
            [
                Batteries.Permissions.CHARGE_ALLOWED,
                Batteries.Permissions.DISCHARGE_ALLOWED,
            ],
        ),
        (
            Batteries.Mode.ZERO_CHARGE_ONLY,
            Batteries.Mode.ZERO,
            [Batteries.Permissions.CHARGE_ALLOWED],
        ),
        (
            Batteries.Mode.ZERO_DISCHARGE_ONLY,
            Batteries.Mode.ZERO,
            [Batteries.Permissions.DISCHARGE_ALLOWED],
        ),
        (Batteries.Mode.TO_FULL, Batteries.Mode.TO_FULL, None),
        (Batteries.Mode.STANDBY, Batteries.Mode.STANDBY, []),
    ],
)
def test_batteries_update_modes_and_permissions(
    mode, expected_mode, expected_permissions
):
    """Test BatteriesUpdate for all modes and permissions."""
    update = BatteriesUpdate.from_mode(mode=mode)

    assert update.mode == expected_mode
    assert update.permissions == expected_permissions


@pytest.mark.parametrize(
    "mode, permissions, expected_mode",
    [
        (
            Batteries.Mode.ZERO,
            [
                Batteries.Permissions.CHARGE_ALLOWED,
                Batteries.Permissions.DISCHARGE_ALLOWED,
            ],
            Batteries.Mode.ZERO,
        ),
        (
            Batteries.Mode.ZERO,
            [Batteries.Permissions.CHARGE_ALLOWED],
            Batteries.Mode.ZERO_CHARGE_ONLY,
        ),
        (
            Batteries.Mode.ZERO,
            [Batteries.Permissions.DISCHARGE_ALLOWED],
            Batteries.Mode.ZERO_DISCHARGE_ONLY,
        ),
        (Batteries.Mode.ZERO, [], Batteries.Mode.STANDBY),
        (Batteries.Mode.STANDBY, [], Batteries.Mode.STANDBY),
        (Batteries.Mode.TO_FULL, [], Batteries.Mode.TO_FULL),
        (Batteries.Mode.ZERO, None, Batteries.Mode.ZERO),
        (Batteries.Mode.STANDBY, None, Batteries.Mode.STANDBY),
        (Batteries.Mode.TO_FULL, None, Batteries.Mode.TO_FULL),
    ],
)
def test_set_mode_based_on_permissions(mode, permissions, expected_mode):
    """Test setting Batteries mode based on permissions."""
    model = Batteries.from_dict(
        {
            "mode": mode,
            "permissions": permissions,
            "power_w": 0.0,
            "target_power_w": 0.0,
            "max_consumption_w": 0.0,
            "max_production_w": 0.0,
        }
    )
    assert model.mode == expected_mode


def test_set_batteries_update_with_invalid_permissions_raises():
    """Test BatteriesUpdate with invalid permissions raises ValueError."""
    with pytest.raises(ValueError):
        BatteriesUpdate.from_mode(
            mode="invalid_mode",  # mypy: ignore [arg-type]
        )



"""Utilities for Python HomeWizard Energy."""

from functools import lru_cache

from awesomeversion import AwesomeVersion


@lru_cache
def get_awesome_version(version: str) -> AwesomeVersion:
    """Return a cached AwesomeVersion object."""
    if version.lower() == "v1":
        return AwesomeVersion("1.0.0")
    return AwesomeVersion(version)



"""Test for Device model."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import Device

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["device"]),
        ("HWE-SKT", ["device"]),
        ("HWE-WTR", ["device"]),
        ("HWE-KWH1", ["device"]),
        ("HWE-KWH3", ["device"]),
        ("SDM230-wifi", ["device"]),
        ("SDM630-wifi", ["device"]),
    ],
)
async def test_device(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Device model."""
    for fixture in fixtures:
        data = Device.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data
        assert snapshot == data


# pylint: disable=too-many-arguments,too-many-positional-arguments
@pytest.mark.parametrize(
    (
        "model",
        "supports_state",
        "supports_identify",
        "supports_cloud_enable",
        "supports_reboot",
        "supports_telegram",
    ),
    [
        ("HWE-P1", False, True, True, False, True),
        ("HWE-SKT", True, True, True, False, False),
        ("HWE-WTR", False, True, True, False, False),
        ("HWE-KWH1", False, False, True, False, False),
        ("HWE-KWH3", False, False, True, False, False),
        ("SDM230-wifi", False, False, True, False, False),
        ("SDM630-wifi", False, False, True, False, False),
    ],
)
async def test_device_support_functions(
    model: str,
    supports_state: bool,
    supports_identify: bool,
    supports_cloud_enable: bool,
    supports_reboot: bool,
    supports_telegram: bool,
):
    """Test Device model support functions."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device
    assert device.supports_state() == supports_state
    assert device.supports_identify() == supports_identify
    assert device.supports_cloud_enable() == supports_cloud_enable
    assert device.supports_reboot() == supports_reboot
    assert device.supports_telegram() == supports_telegram



"""Test for Measurement model."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import Measurement

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-P1",
            [
                "measurement_1_phase_no_gas",
                "measurement_3_phase_with_gas_with_watermeter",
                "measurement_invalid_ean",
                "measurement_invalid_external",
            ],
        ),
        ("HWE-KWH1", ["measurement"]),
        ("HWE-KWH3", ["measurement"]),
        ("HWE-BAT", ["measurement"]),
    ],
)
async def test_measurement(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Measurement model."""
    for fixture in fixtures:
        data = Measurement.from_dict(
            json.loads(load_fixtures(f"{model}/{fixture}.json"))
        )
        assert data
        assert snapshot == data


async def test_measurement_ignores_invalid_tariff():
    """Test Measurement model ignores invalid tariff."""
    measurement = Measurement.from_dict({"tariff": 5432})
    assert measurement
    assert measurement.tariff is None



"""Test for State model."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import State

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-SKT",
            ["state_all", "state_power_on", "state_switch_lock", "state_brightness"],
        ),
    ],
)
async def test_state(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test State model."""
    for fixture in fixtures:
        data = State.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data
        assert snapshot == data



"""Representation of a HomeWizard Energy device."""

from __future__ import annotations

import asyncio
import json
import ssl
from collections.abc import Callable, Coroutine
from http import HTTPStatus
from typing import Any, TypeVar

import async_timeout
import backoff
from aiohttp.client import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_DELETE, METH_GET, METH_POST, METH_PUT
from mashumaro.exceptions import InvalidFieldValue, MissingField

from ..const import LOGGER
from ..errors import (
    DisabledError,
    InvalidUserNameError,
    NotFoundError,
    RequestError,
    ResponseError,
    UnauthorizedError,
    UnsupportedError,
)
from ..homewizard_energy import HomeWizardEnergy
from ..models import (
    Batteries,
    BatteriesUpdate,
    Device,
    Measurement,
    System,
    SystemUpdate,
    Token,
)
from .cacert import CACERT

T = TypeVar("T")


def authorized_method(
    func: Callable[..., Coroutine[Any, Any, T]],
) -> Callable[..., Coroutine[Any, Any, T]]:
    """Decorator method to check if token is set."""

    async def wrapper(self, *args, **kwargs) -> T:
        # pylint: disable=protected-access
        if self._token is None:
            raise UnauthorizedError("Token missing")

        return await func(self, *args, **kwargs)

    return wrapper


# pylint: disable=abstract-method
class HomeWizardEnergyV2(HomeWizardEnergy):
    """Communicate with a HomeWizard Energy device."""

    _ssl: ssl.SSLContext | bool = False
    _identifier: str | None = None

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        host: str,
        identifier: str | None = None,
        token: str | None = None,
        clientsession: ClientSession = None,
        timeout: int = 10,
    ):
        """Create a HomeWizard Energy object.

        Args:
            host: IP or URL for device.
            id: ID for device.
            token: Token for device.
            timeout: Request timeout in seconds.
        """
        super().__init__(host, clientsession, timeout)
        self._identifier = identifier
        self._token = token

    @authorized_method
    async def device(self, reset_cache: bool = False) -> Device:
        """Return the device object."""
        if self._device is not None and not reset_cache:
            return self._device

        _, response = await self._request("/api")
        device = Device.from_json(response)

        # Cache device object
        self._device = device
        return device

    @authorized_method
    async def measurement(self) -> Measurement:
        """Return the measurement object."""
        _, response = await self._request("/api/measurement")
        measurement = Measurement.from_json(response)

        return measurement

    @authorized_method
    async def telegram(self) -> str:
        """Return the most recent, valid telegram that was given by the device.
        The telegram is not processed in any other form.
        If you need parsed data, use the measurement method.
        """
        _, telegram = await self._request("/api/telegram")
        return telegram

    @authorized_method
    async def system(
        self,
        cloud_enabled: bool | None = None,
        status_led_brightness_pct: int | None = None,
        api_v1_enabled: bool | None = None,
    ) -> System:
        """Return the system object."""

        if (
            cloud_enabled is not None
            or status_led_brightness_pct is not None
            or api_v1_enabled is not None
        ):
            data = SystemUpdate(
                cloud_enabled=cloud_enabled,
                status_led_brightness_pct=status_led_brightness_pct,
                api_v1_enabled=api_v1_enabled,
            ).to_dict()
            status, response = await self._request(
                "/api/system", method=METH_PUT, data=data
            )

        else:
            status, response = await self._request("/api/system")

        if status != HTTPStatus.OK:
            error = json.loads(response).get("error", response)
            raise RequestError(f"Failed to get system: {error}")

        system = System.from_json(response)
        return system

    @authorized_method
    async def batteries(
        self,
        mode: Batteries.Mode | None = None,
    ) -> Batteries:
        """Return the batteries object."""

        if self._device is not None and self._device.supports_batteries() is False:
            raise UnsupportedError("Batteries is not supported")

        try:
            if mode is not None:
                data = BatteriesUpdate.from_mode(mode).to_dict()
                status, response = await self._request(
                    "/api/batteries", method=METH_PUT, data=data
                )
            else:
                status, response = await self._request("/api/batteries")

            if status != HTTPStatus.OK:
                error = json.loads(response).get("error", response)
                raise RequestError(f"Failed to get batteries: {error}")
        except NotFoundError as exception:
            # The batteries endpoint is not available on the device
            raise UnsupportedError("Batteries is not supported") from exception

        return Batteries.from_json(response)

    @authorized_method
    async def identify(
        self,
    ) -> None:
        """Send identify request."""
        await self._request("/api/system/identify", method=METH_PUT)

    @authorized_method
    async def reboot(
        self,
    ) -> None:
        """Reboot the HomeWizard Energy device.

        This will cause the device to restart, resulting in temporary unavailability.
        The reboot process typically takes a few seconds to complete.

        Note: A reboot is usually not necessary.
        Make sure to inform the user that if the issue persists and frequent reboots are required,
        they need to contact our support team to help identify and resolve the root cause.
        """
        await self._request("/api/system/reboot", method=METH_PUT)

    async def get_token(
        self,
        name: str,
    ) -> str:
        """Get authorization token from device."""
        status, response = await self._request(
            "/api/user", method=METH_POST, data={"name": f"local/{name}"}
        )

        if status == HTTPStatus.FORBIDDEN:
            raise DisabledError("User creation is not enabled on the device")

        if status != HTTPStatus.OK:
            error = json.loads(response).get("error", response)
            raise InvalidUserNameError(
                f"Error occurred while getting token: {error}", error
            )

        try:
            token = Token.from_json(response).token
        except (InvalidFieldValue, MissingField) as ex:
            raise ResponseError("Failed to get token") from ex

        self._token = token
        return token

    @authorized_method
    async def delete_token(
        self,
        name: str | None = None,
    ) -> None:
        """Delete authorization token from device."""
        status, response = await self._request(
            "/api/user",
            method=METH_DELETE,
            data={"name": name} if name is not None else None,
        )

        if status != HTTPStatus.NO_CONTENT:
            error = json.loads(response).get("error", response)
            raise RequestError(f"Error occurred while getting token: {error}", error)

        # Our token was invalided, resetting it
        if name is None:
            self._token = None

    async def _get_ssl_context(self) -> ssl.SSLContext:
        """
        Get a clientsession that is tuned for communication with the HomeWizard Energy Device
        """

        def _build_ssl_context() -> ssl.SSLContext:
            context = ssl.create_default_context(cadata=CACERT)
            context.verify_flags = ssl.VERIFY_X509_PARTIAL_CHAIN  # pylint: disable=no-member
            if self._identifier is not None:
                context.hostname_checks_common_name = True
            else:
                context.check_hostname = False  # Skip hostname validation
                context.verify_mode = ssl.CERT_REQUIRED  # Keep SSL verification active
            return context

        # Creating an SSL context has some blocking IO so need to run it in the executor
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, _build_ssl_context)

    @backoff.on_exception(backoff.expo, RequestError, max_tries=3, logger=None)
    async def _request(
        self, path: str, method: str = METH_GET, data: object = None
    ) -> tuple[HTTPStatus, dict[str, Any] | None]:
        """Make a request to the API."""

        async with self._lock:
            if self._session is None:
                await self._create_clientsession()

        async with self._lock:
            if self._ssl is False:
                self._ssl = await self._get_ssl_context()

        # Construct request
        url = f"https://{self.host}{path}"
        headers = {
            "Content-Type": "application/json",
        }
        if self._token is not None:
            headers["Authorization"] = f"Bearer {self._token}"

        LOGGER.debug("%s, %s, %s", method, url, data)

        try:
            async with async_timeout.timeout(self._request_timeout):
                async with self._lock:
                    resp = await self._session.request(
                        method,
                        url,
                        json=data,
                        headers=headers,
                        ssl=self._ssl,
                        server_hostname=self._identifier,
                    )
                LOGGER.debug("%s, %s", resp.status, await resp.text("utf-8"))
        except asyncio.TimeoutError as exception:
            raise RequestError(
                f"Timeout occurred while connecting to the HomeWizard Energy device at {self.host}"
            ) from exception
        except (ClientError, ClientResponseError) as exception:
            raise RequestError(
                f"Error occurred while communicating with the HomeWizard Energy device at {self.host}"
            ) from exception

        match resp.status:
            case HTTPStatus.UNAUTHORIZED:
                raise UnauthorizedError("Token rejected")
            case HTTPStatus.NO_CONTENT:
                # No content, just return
                return (HTTPStatus.NO_CONTENT, None)
            case HTTPStatus.NOT_FOUND:
                raise NotFoundError("Resource not found")
            case HTTPStatus.OK:
                pass

        return (resp.status, await resp.text())

    async def __aenter__(self) -> HomeWizardEnergyV2:
        """Async enter.

        Returns:
            The HomeWizardEnergyV2 object.
        """
        return self



"""Test for CombinedModels model."""

import json

import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.models import CombinedModels, Device, Measurement, State

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model"),
    [
        "HWE-P1",
        "HWE-SKT",
        "HWE-WTR",
        "HWE-KWH1",
        "HWE-KWH3",
        "SDM230-wifi",
        "SDM630-wifi",
    ],
)
async def test_combined_model(model: str, snapshot: SnapshotAssertion):
    """Test combined model."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    measurement = Measurement.from_dict(json.loads(load_fixtures(f"{model}/data.json")))
    combined = CombinedModels(
        device=device, measurement=measurement, state=None, system=None
    )
    assert snapshot == combined


async def test_combined_remaps_legacy_wifi_ssid_to_system(snapshot: SnapshotAssertion):
    """Test CombinedModels model remaps wifi_ssid to system."""
    device = Device.from_dict(json.loads(load_fixtures("HWE-SKT/device.json")))
    measurement = Measurement.from_dict(json.loads(load_fixtures("HWE-SKT/data.json")))
    combined = CombinedModels(
        device=device, measurement=measurement, state=None, system=None
    )
    assert combined.system.wifi_ssid == measurement.wifi_ssid
    assert snapshot == combined


async def test_combined_remaps_legacy_brightness_to_system(snapshot: SnapshotAssertion):
    """Test CombinedModels model remaps wifi_ssid to system."""
    device = Device.from_dict(json.loads(load_fixtures("HWE-SKT/device.json")))
    state = State.from_dict(json.loads(load_fixtures("HWE-SKT/state_all.json")))
    combined = CombinedModels(device=device, measurement=None, state=state, system=None)
    assert combined.system.status_led_brightness_pct == state.brightness / 2.55
    assert snapshot == combined



"""Test for HomeWizard Energy."""

import asyncio
from unittest.mock import AsyncMock

import aiohttp
import pytest
from syrupy.assertion import SnapshotAssertion

from homewizard_energy import HomeWizardEnergyV2
from homewizard_energy.errors import (
    DisabledError,
    InvalidUserNameError,
    RequestError,
    ResponseError,
    UnauthorizedError,
)
from homewizard_energy.models import Batteries

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]

### CombinedModels tests ###


async def test_combined_models_without_authentication():
    """Test combined models request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.combined()


async def test_combined_models_with_invalid_authentication(aresponses):
    """Test combined models request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.combined()


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@pytest.mark.parametrize(
    ("model", "device", "measurement", "state", "system", "batteries"),
    [
        ("HWE-P1", "device", "measurement_1_phase_no_gas", None, "system", "batteries"),
        ("HWE-KWH1", "device", "measurement", None, "system", "batteries"),
        ("HWE-KWH3", "device", "measurement", None, "system", "batteries"),
        ("HWE-BAT", "device", "measurement", None, "system", None),
    ],
)
async def test_combined_models_with_valid_authentication(
    model: str,
    device: str,
    measurement: str,
    state: str | None,
    system: str,
    batteries: str | None,
    snapshot: SnapshotAssertion,
    aresponses,
):
    """Test combined models request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{device}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{measurement}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )
    aresponses.add(
        "example.com",
        "/api/batteries",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{batteries}.json")
            if batteries
            else "404 Not Found",
            status=200 if batteries else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        data = await api.combined()
        assert data is not None
        assert data == snapshot


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@pytest.mark.parametrize(
    ("model", "device", "measurement", "state", "system", "batteries"),
    [
        ("HWE-P1", "device", "measurement_1_phase_no_gas", None, "system", "batteries"),
        ("HWE-KWH1", "device", "measurement", None, "system", "batteries"),
        ("HWE-KWH3", "device", "measurement", None, "system", "batteries"),
        ("HWE-BAT", "device", "measurement", None, "system", None),
    ],
)
async def test_combined_models_with_valid_authentication_caches_device(
    model: str,
    device: str,
    measurement: str,
    state: str | None,
    system: str,
    batteries: str | None,
    snapshot: SnapshotAssertion,
    aresponses,
):
    """Test combined models request is successful when valid authentication is provided."""

    # Request 1
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{device}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{measurement}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )
    aresponses.add(
        "example.com",
        "/api/batteries",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{batteries}.json")
            if batteries
            else "404 Not Found",
            status=200 if batteries else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    # Request 2, should use cache of `/api`
    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{measurement}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/state",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{state}.json") if state else "404 Not Found",
            status=200 if state else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/batteries",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{batteries}.json")
            if batteries
            else "404 Not Found",
            status=200 if batteries else 404,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        data = await api.combined()
        assert data is not None
        assert data == snapshot

        data_2 = await api.combined()
        assert data_2 == data


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@pytest.mark.parametrize(
    ("model", "device", "measurement", "system"),
    [
        ("HWE-BAT", "device", "measurement", "system"),
    ],
)
async def test_batteries_request_is_not_requested_when_known_unsupported(
    model: str,
    device: str,
    measurement: str,
    system: str,
    aresponses,
):
    """Test combined models request is does not request unsupported API's when already known it is unsupported."""

    # Request 1
    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{device}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{measurement}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            text=load_fixtures(f"{model}/{system}.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        await api.device()
        await api.combined()


### Device tests ###


async def test_class_returns_host():
    """Test class returns host."""

    async with HomeWizardEnergyV2("example.com") as api:
        assert api.host == "example.com"


async def test_device_without_authentication():
    """Test device request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.device()


async def test_device_with_invalid_authentication(aresponses):
    """Test device request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.device()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["device"]),
    ],
)
async def test_device_with_valid_authentication(
    model: str, fixtures: list[str], snapshot: SnapshotAssertion, aresponses
):
    """Test device request is successful when valid authentication is provided."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json"},
            ),
        )

        async with HomeWizardEnergyV2("example.com", token="token") as api:
            device = await api.device()

            assert device
            assert device.product_type == model

            assert device == snapshot


### Measurement tests ###


async def test_measurement_without_authentication():
    """Test measurement request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.measurement()


async def test_measurement_with_invalid_authentication(aresponses):
    """Test measurement request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/measurement",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.measurement()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        (
            "HWE-P1",
            [
                "measurement_1_phase_no_gas",
                "measurement_3_phase_with_gas_with_watermeter",
            ],
        ),
    ],
)
async def test_measurement_with_valid_authentication(
    model: str, fixtures: list[str], snapshot: SnapshotAssertion, aresponses
):
    """Test measurement request is successful when valid authentication is provided."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api/measurement",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json"},
            ),
        )

        async with HomeWizardEnergyV2("example.com", token="token") as api:
            measurement = await api.measurement()
            assert measurement is not None
            assert measurement == snapshot


### Telegram tests ###


async def test_telegram_without_authentication():
    """Test telegram request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.telegram()


async def test_telegram_with_invalid_authentication(aresponses):
    """Test telegram request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/telegram",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/text"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.telegram()


async def test_telegram_with_valid_authentication(
    snapshot: SnapshotAssertion, aresponses
):
    """Test telegram request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/telegram",
        "GET",
        aresponses.Response(
            text=load_fixtures("HWE-P1/telegram.txt"),
            status=200,
            headers={"Content-Type": "application/txt"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        telegram = await api.telegram()
        assert telegram is not None
        assert telegram == snapshot


### System tests ###


async def test_system_without_authentication():
    """Test system request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.system()


async def test_system_with_invalid_authentication(aresponses):
    """Test system request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.system()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["system"]),
    ],
)
async def test_system_with_valid_authentication(
    model: str, fixtures: list[str], snapshot: SnapshotAssertion, aresponses
):
    """Test system request is successful when valid authentication is provided."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api/system",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json"},
            ),
        )

        async with HomeWizardEnergyV2("example.com", token="token") as api:
            system = await api.system()
            assert system is not None
            assert system == snapshot


async def test_system_set_with_valid_authentication(aresponses):
    """Test system set request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system",
        "PUT",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"cloud_enabled": true}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        data = await api.system(cloud_enabled=True)
        assert data is not None
        assert data.cloud_enabled is True


async def test_system_returns_unexpected_response(aresponses):
    """Test system set request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text='{"error": "server:error"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(RequestError) as e:
            await api.system()
            assert str(e.value) == "server:error"


### Batteries tests ###


async def test_batteries_without_authentication():
    """Test batteries request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.batteries()


async def test_batteries_with_invalid_authentication(aresponses):
    """Test batteries request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/batteries",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.batteries()


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["batteries"]),
    ],
)
async def test_batteries_with_valid_authentication(
    model: str, fixtures: list[str], snapshot: SnapshotAssertion, aresponses
):
    """Test batteries request is successful when valid authentication is provided."""

    for fixture in fixtures:
        aresponses.add(
            "example.com",
            "/api/batteries",
            "GET",
            aresponses.Response(
                text=load_fixtures(f"{model}/{fixture}.json"),
                status=200,
                headers={"Content-Type": "application/json"},
            ),
        )

        async with HomeWizardEnergyV2("example.com", token="token") as api:
            batteries = await api.batteries()
            assert batteries is not None
            assert batteries == snapshot


async def test_batteries_returns_unexpected_response(aresponses):
    """Test batteries request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/batteries",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text='{"error": "server:error"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(RequestError) as e:
            await api.batteries()
            assert str(e.value) == "server:error"


async def test_batteries_put_without_authentication():
    """Test batteries request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.batteries(mode=Batteries.Mode.STANDBY)


async def test_batteries_put_with_invalid_authentication(aresponses):
    """Test batteries request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/batteries",
        "PUT",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.batteries(mode=Batteries.Mode.STANDBY)


@pytest.mark.parametrize(
    ("model"),
    [
        ("HWE-P1"),
    ],
)
async def test_batteries_put_with_valid_authentication(
    model: str, snapshot: SnapshotAssertion, aresponses
):
    """Test batteries request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/batteries",
        "PUT",
        aresponses.Response(
            text=load_fixtures(f"{model}/batteries.json"),
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        batteries = await api.batteries(mode=Batteries.Mode.STANDBY)
        assert batteries is not None
        assert batteries == snapshot


### Identify tests ###


async def test_identify_without_authentication():
    """Test identify request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.identify()


async def test_identify_with_invalid_authentication(aresponses):
    """Test identify request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system/identify",
        "PUT",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.identify()


async def test_identify_with_authentication(aresponses):
    """Test identify request is successful when authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system/identify",
        "PUT",
        aresponses.Response(
            status=204,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        data = await api.identify()
        assert data is None


### Reboot tests ###


async def test_reboot_without_authentication():
    """Test reboot request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.reboot()


async def test_reboot_with_invalid_authentication(aresponses):
    """Test reboot request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system/reboot",
        "PUT",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.reboot()


async def test_reboot_with_authentication(aresponses):
    """Test reboot request is successful when authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/system/reboot",
        "PUT",
        aresponses.Response(
            status=204,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        data = await api.reboot()
        assert data is None


### Get token tests ###


async def test_get_token_without_permission(aresponses):
    """Test get token request is rejected when no permission, user must authorize."""

    aresponses.add(
        "example.com",
        "/api/user",
        "POST",
        aresponses.Response(
            status=403,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:creation-not-enabled"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(DisabledError):
            await api.get_token("name")


async def test_get_token_with_permission(aresponses):
    """Test get token request accepted."""

    aresponses.add(
        "example.com",
        "/api/user",
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"token": "new_token", "name": "local/name"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com") as api:
        token = await api.get_token("name")
        assert token == "new_token"


async def test_get_token_returns_unexpected_response_code(aresponses):
    """Test get token request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "POST",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text='{"error": "server:error"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(InvalidUserNameError) as e:
            await api.get_token("name")
            assert str(e.value) == "server:error"


async def test_get_token_returns_unexpected_response_data(aresponses):
    """Test get token request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:error"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(ResponseError) as e:
            await api.get_token("name")
            assert str(e.value) == "user:error"


async def test_delete_token_without_authentication():
    """Test delete token request is rejected when no authentication is provided."""

    async with HomeWizardEnergyV2("example.com") as api:
        with pytest.raises(UnauthorizedError):
            await api.delete_token()


async def test_delete_token_with_invalid_authentication(aresponses):
    """Test delete token request is unsuccessful when invalid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "DELETE",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text='{"error": "user:unauthorized"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(UnauthorizedError):
            await api.delete_token()


async def test_delete_token_with_authentication(aresponses):
    """Test delete_token request is successful when authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "DELETE",
        aresponses.Response(
            status=204,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        await api.delete_token()
        assert api._token is None  # pylint: disable=protected-access


async def test_delete_token_from_other_user_with_authentication(aresponses):
    """Test delete_token request is successful when authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "DELETE",
        aresponses.Response(
            status=204,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        await api.delete_token(name="local/other")
        assert api._token == "token"  # pylint: disable=protected-access


async def test_delete_token_returns_unexpected_response_code(aresponses):
    """Test delete token request is successful when valid authentication is provided."""

    aresponses.add(
        "example.com",
        "/api/user",
        "DELETE",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text='{"error": "server:error"}',
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token") as api:
        with pytest.raises(RequestError) as e:
            await api.delete_token()
            assert str(e.value) == "server:error"


### Generic request tests ###


# pylint: disable=protected-access
async def test_request_handles_timeout():
    """Test request times out when request takes too long."""
    async with HomeWizardEnergyV2("example.com", token="token") as api:
        api._session = AsyncMock()
        api._session.request = AsyncMock(side_effect=asyncio.TimeoutError)

        with pytest.raises(RequestError):
            await api.device()

        assert api._session.request.call_count == 3


# pylint: disable=protected-access
async def test_request_handles_clienterror():
    """Test request times out when request takes too long."""
    async with HomeWizardEnergyV2("example.com", token="token") as api:
        api._session = AsyncMock()
        api._session.request = AsyncMock(side_effect=aiohttp.ClientError)

        with pytest.raises(RequestError):
            await api.device()

        assert api._session.request.call_count == 3


async def test_request_with_identifier_sets_common_name(aresponses):
    """Test request with identifier sets common name."""

    device = load_fixtures("HWE-P1/device.json")
    device.replace("HWE-P1", "NEW-DEVICE")

    aresponses.add(
        "example.com",
        "/api",
        "GET",
        aresponses.Response(
            text=device,
            status=200,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with HomeWizardEnergyV2("example.com", token="token", identifier="id") as api:
        data = await api.device()
        assert data is not None



"""Tests for python-homewizard-energy."""

import os


def load_fixtures(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path, encoding="utf-8") as fptr:
        return fptr.read()



"""Test for Device model."""

import json

import pytest
from awesomeversion import AwesomeVersion
from syrupy.assertion import SnapshotAssertion

from homewizard_energy.const import Model
from homewizard_energy.models import Batteries, Device

from . import load_fixtures

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "fixtures"),
    [
        ("HWE-P1", ["device"]),
        ("HWE-KWH1", ["device"]),
        ("HWE-KWH3", ["device"]),
        ("HWE-BAT", ["device"]),
    ],
)
async def test_device(model: str, fixtures: str, snapshot: SnapshotAssertion):
    """Test Device model."""
    for fixture in fixtures:
        data = Device.from_dict(json.loads(load_fixtures(f"{model}/{fixture}.json")))
        assert data
        assert snapshot == data


# pylint: disable=too-many-arguments,too-many-positional-arguments
@pytest.mark.parametrize(
    (
        "model",
        "supports_state",
        "supports_identify",
        "supports_cloud_enable",
        "supports_reboot",
        "supports_telegram",
    ),
    [
        ("HWE-P1", False, True, True, True, True),
        ("HWE-KWH1", False, False, True, True, False),
        ("HWE-KWH3", False, False, True, True, False),
        ("HWE-BAT", False, True, False, False, False),
    ],
)
async def test_device_support_functions(
    model: str,
    supports_state: bool,
    supports_identify: bool,
    supports_cloud_enable: bool,
    supports_reboot: bool,
    supports_telegram: bool,
):
    """Test Device model support functions."""
    device = Device.from_dict(json.loads(load_fixtures(f"{model}/device.json")))
    assert device
    assert device.supports_state() == supports_state
    assert device.supports_identify() == supports_identify
    assert device.supports_cloud_enable() == supports_cloud_enable
    assert device.supports_reboot() == supports_reboot
    assert device.supports_telegram() == supports_telegram


@pytest.mark.parametrize(
    "product_type,api_version,expected_modes",
    [  # Devices that do not support batteries
        (Model.ENERGY_SOCKET, "2.2.0", None),
        (Model.WATER_METER, "2.2.0", None),
        # Devices that support batteries, API < 2.2.0
        (
            Model.P1_METER,
            "2.1.0",
            [Batteries.Mode.ZERO, Batteries.Mode.TO_FULL, Batteries.Mode.STANDBY],
        ),
        (
            Model.ENERGY_METER_1_PHASE,
            "2.1.0",
            [Batteries.Mode.ZERO, Batteries.Mode.TO_FULL, Batteries.Mode.STANDBY],
        ),
        (
            Model.ENERGY_METER_3_PHASE,
            "2.1.0",
            [Batteries.Mode.ZERO, Batteries.Mode.TO_FULL, Batteries.Mode.STANDBY],
        ),
        (
            Model.ENERGY_METER_EASTRON_SDM230,
            "2.1.0",
            [Batteries.Mode.ZERO, Batteries.Mode.TO_FULL, Batteries.Mode.STANDBY],
        ),
        (
            Model.ENERGY_METER_EASTRON_SDM630,
            "2.1.0",
            [Batteries.Mode.ZERO, Batteries.Mode.TO_FULL, Batteries.Mode.STANDBY],
        ),
        # Devices that support batteries, API >= 2.2.0
        (
            Model.P1_METER,
            "2.2.0",
            [
                Batteries.Mode.ZERO,
                Batteries.Mode.TO_FULL,
                Batteries.Mode.STANDBY,
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ],
        ),
        (
            Model.ENERGY_METER_1_PHASE,
            "2.2.0",
            [
                Batteries.Mode.ZERO,
                Batteries.Mode.TO_FULL,
                Batteries.Mode.STANDBY,
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ],
        ),
        (
            Model.ENERGY_METER_3_PHASE,
            "2.2.0",
            [
                Batteries.Mode.ZERO,
                Batteries.Mode.TO_FULL,
                Batteries.Mode.STANDBY,
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ],
        ),
        (
            Model.ENERGY_METER_EASTRON_SDM230,
            "2.2.0",
            [
                Batteries.Mode.ZERO,
                Batteries.Mode.TO_FULL,
                Batteries.Mode.STANDBY,
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ],
        ),
        (
            Model.ENERGY_METER_EASTRON_SDM630,
            "2.2.0",
            [
                Batteries.Mode.ZERO,
                Batteries.Mode.TO_FULL,
                Batteries.Mode.STANDBY,
                Batteries.Mode.ZERO_CHARGE_ONLY,
                Batteries.Mode.ZERO_DISCHARGE_ONLY,
            ],
        ),
    ],
)
def test_supported_battery_modes(product_type, api_version, expected_modes):
    """Test supported_battery_modes method of Device."""
    device = Device(
        product_name="Test Device",
        product_type=product_type,
        serial="1234567890",
        api_version=AwesomeVersion(api_version),
        firmware_version="1.0.0",
    )
    result = device.supported_battery_modes()
    if expected_modes is None:
        assert result is None
    else:
        assert set(result) == set(
            expected_modes
        )  # We don't care about order, just presence



"""Representation of a HomeWizard Energy device."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine
from http import HTTPStatus
from typing import Any, TypeVar

import async_timeout
import backoff
from aiohttp.client import ClientError, ClientResponseError
from aiohttp.hdrs import METH_GET, METH_PUT

from ..const import LOGGER
from ..errors import DisabledError, NotFoundError, RequestError, UnsupportedError
from ..homewizard_energy import HomeWizardEnergy
from ..models import Device, Measurement, State, StateUpdate, System, SystemUpdate

T = TypeVar("T")


def optional_method(
    func: Callable[..., Coroutine[Any, Any, T]],
) -> Callable[..., Coroutine[Any, Any, T]]:
    """Check if method is supported."""

    async def wrapper(self, *args, **kwargs) -> T:
        try:
            return await func(self, *args, **kwargs)
        except NotFoundError as ex:
            raise UnsupportedError(f"{func.__name__} is not supported") from ex

    return wrapper


# pylint: disable=abstract-method
class HomeWizardEnergyV1(HomeWizardEnergy):
    """Communicate with a HomeWizard Energy device."""

    async def device(self, reset_cache: bool = False) -> Device:
        """Return the device object."""

        if self._device is not None and not reset_cache:
            return self._device

        _, response = await self._request("api")
        device = Device.from_json(response)

        # Cache device object
        self._device = device
        return device

    async def measurement(self) -> Measurement:
        """Return the data object."""
        _, response = await self._request("api/v1/data")
        return Measurement.from_json(response)

    @optional_method
    async def telegram(self) -> str:
        """Return the most recent, valid telegram that was given by the device.
        The telegram is not processed in any other form.
        If you need parsed data, use the measurement method.
        """
        if self._device is not None and self._device.supports_telegram() is False:
            raise UnsupportedError("Telegram is not supported")

        _, telegram = await self._request("api/v1/telegram")
        return telegram

    @optional_method
    async def system(
        self,
        cloud_enabled: bool | None = None,
        status_led_brightness_pct: int | None = None,
        api_v1_enabled: bool | None = None,
    ) -> System:
        """Return the system object."""

        # Legacy: raise on unsupported field
        if api_v1_enabled is not None:
            raise UnsupportedError("Setting api_v1_enabled is not supported in v1")

        # Legacy: route 'status_led_brightness_pct' to state
        if status_led_brightness_pct is not None:
            state = await self.state(brightness=status_led_brightness_pct * 2.55)
            return System(status_led_brightness_pct=state.brightness / 2.55)

        if cloud_enabled is not None:
            # Executing the update
            data = SystemUpdate(cloud_enabled=cloud_enabled).to_dict()
            _, response = await self._request(
                "api/v1/system", method=METH_PUT, data=data
            )

        else:
            _, response = await self._request("api/v1/system")

        system = System.from_json(response)
        return system

    @optional_method
    async def state(
        self,
        power_on: bool | None = None,
        switch_lock: bool | None = None,
        brightness: int | None = None,
    ) -> State:
        """Return or update the state object."""

        if self._device is not None and self._device.supports_state() is False:
            raise UnsupportedError("State is not supported")

        if power_on is not None or switch_lock is not None or brightness is not None:
            data = StateUpdate(
                power_on=power_on, switch_lock=switch_lock, brightness=brightness
            ).to_dict()
            _, response = await self._request(
                "api/v1/state", method=METH_PUT, data=data
            )

        else:
            _, response = await self._request("api/v1/state")

        state = State.from_json(response)
        return state

    @optional_method
    async def identify(
        self,
    ) -> bool:
        """Send identify request."""

        if self._device is not None and self._device.supports_identify() is False:
            raise UnsupportedError("State is not supported")

        await self._request("api/v1/identify", method=METH_PUT)
        return True

    @backoff.on_exception(backoff.expo, RequestError, max_tries=3, logger=None)
    async def _request(
        self, path: str, method: str = METH_GET, data: object = None
    ) -> tuple[HTTPStatus, dict[str, Any] | None]:
        """Make a request to the API."""

        if self._session is None:
            await self._create_clientsession()

        # Construct request
        url = f"http://{self.host}/{path}"
        headers = {"Content-Type": "application/json"}

        LOGGER.debug("%s, %s, %s", method, url, data)

        try:
            async with async_timeout.timeout(self._request_timeout):
                resp = await self._session.request(
                    method,
                    url,
                    json=data,
                    headers=headers,
                )
                LOGGER.debug("%s, %s", resp.status, await resp.text("utf-8"))
        except asyncio.TimeoutError as exception:
            raise RequestError(
                f"Timeout occurred while connecting to the HomeWizard Energy device at {self.host}"
            ) from exception
        except (ClientError, ClientResponseError) as exception:
            raise RequestError(
                f"Error occurred while communicating with the HomeWizard Energy device at {self.host}"
            ) from exception

        match resp.status:
            case HTTPStatus.FORBIDDEN:
                raise DisabledError(
                    "API disabled. API must be enabled in HomeWizard Energy app"
                )
            case HTTPStatus.NOT_FOUND:
                raise NotFoundError("Resource not found")
            case HTTPStatus.OK:
                pass

        if resp.status != HTTPStatus.OK:
            # Something else went wrong
            raise RequestError(f"API request error ({resp.status})")

        return (resp.status, await resp.text())

    async def __aenter__(self) -> HomeWizardEnergyV1:
        """Async enter.

        Returns:
            The HomeWizardEnergyV1 object.
        """
        return self



"""HomeWizard Energy base class"""

from __future__ import annotations

import asyncio
from typing import Any

from aiohttp.client import ClientSession, ClientTimeout, TCPConnector

from .const import LOGGER
from .errors import UnsupportedError
from .models import Batteries, CombinedModels, Device, Measurement, State, System


class HomeWizardEnergy:
    """Base class for HomeWizard Energy API."""

    _session: ClientSession | None = None
    _close_session: bool = False
    _request_timeout: int = 10
    _host: str

    _device: Device | None = None

    _lock: asyncio.Lock

    def __init__(
        self,
        host: str,
        clientsession: ClientSession = None,
        timeout: int = 10,
    ):
        """Create a HomeWizard Energy object.

        Args:
            host: IP or URL for device.
            clientsession: The clientsession.
            timeout: Request timeout in seconds.
        """
        self._host = host
        self._session = clientsession
        self._close_session = clientsession is None
        self._request_timeout = timeout

        self._lock = asyncio.Lock()

    @property
    def host(self) -> str:
        """Return the hostname of the device.

        Returns:
            host: The used host

        """
        return self._host

    async def combined(self) -> CombinedModels:
        """Get all information."""

        async def fetch_data(coroutine):
            try:
                return await coroutine
            except (UnsupportedError, NotImplementedError):
                return None

        device, measurement, system, state, batteries = await asyncio.gather(
            fetch_data(self.device()),
            fetch_data(self.measurement()),
            fetch_data(self.system()),
            fetch_data(self.state()),
            fetch_data(self.batteries()),
        )

        return CombinedModels(
            device=device,
            measurement=measurement,
            system=system,
            state=state,
            batteries=batteries,
        )

    async def device(self, reset_cache: bool = False) -> Device:
        """Get the device information."""
        raise NotImplementedError

    async def measurement(self) -> Measurement:
        """Get the current measurement."""
        raise NotImplementedError

    async def telegram(self) -> str:
        """Get the latest telegram."""
        raise NotImplementedError

    async def system(
        self,
        cloud_enabled: bool | None = None,
        status_led_brightness_pct: int | None = None,
        api_v1_enabled: bool | None = None,
    ) -> System:
        """Get/set the system."""
        raise NotImplementedError

    async def state(
        self,
        power_on: bool | None = None,
        switch_lock: bool | None = None,
        brightness: int | None = None,
    ) -> State:
        """Get/set the state."""
        raise UnsupportedError("State is not supported")

    async def batteries(self, mode: Batteries.Mode | None = None) -> Batteries:
        """Get the batteries."""
        raise UnsupportedError("Batteries are not supported")

    async def identify(
        self,
    ) -> None:
        """Identify the device."""
        raise NotImplementedError

    async def reboot(
        self,
    ) -> None:
        """Reboot the device."""
        raise UnsupportedError("Reboot is not supported")

    async def close(self) -> None:
        """Close client session."""
        LOGGER.debug("Closing clientsession")
        if self._session and self._close_session:
            await self._session.close()

    async def _create_clientsession(self) -> None:
        """Create a client session."""

        LOGGER.debug("Creating clientsession")

        if self._session is not None:
            raise RuntimeError("Session already exists")  # pragma: no cover

        connector = TCPConnector(
            enable_cleanup_closed=True,
            limit_per_host=1,
        )

        self._close_session = True

        self._session = ClientSession(
            connector=connector, timeout=ClientTimeout(total=self._request_timeout)
        )

    async def __aenter__(self) -> HomeWizardEnergy:
        """Async enter.

        Returns:
            The HomeWizardEnergy object.
        """
        return self

    async def __aexit__(self, *_exc_info: Any) -> None:
        """Async exit.

        Args:
            _exc_info: Exec type.
        """
        await self.close()



"""Test the helper functions."""

import pytest

from homewizard_energy.models import get_verification_hostname

pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize(
    ("model", "expected"),
    [
        ("HWE-P1", "appliance/p1dongle/1234567890"),
        ("HWE-SKT", "appliance/energysocket/1234567890"),
        ("HWE-WTR", "appliance/watermeter/1234567890"),
        ("HWE-DSP", "appliance/display/1234567890"),
        ("HWE-KWH1", "appliance/energymeter/1234567890"),
        ("SDM230-wifi", "appliance/energymeter/1234567890"),
        ("HWE-KWH3", "appliance/energymeter/1234567890"),
        ("SDM630-wifi", "appliance/energymeter/1234567890"),
        ("HWE-BAT", "appliance/battery/1234567890"),
    ],
)
async def test_get_verification_hostname_returns_expected_identifier(
    model: str, expected: str
):
    """Test if get_verification_hostname returns the expected identifier."""
    result = get_verification_hostname(model, "1234567890")
    assert result == expected


async def test_get_verification_hostname_raises_value_error():
    """Test if get_verification_hostname raises a ValueError for an unsupported model."""
    with pytest.raises(ValueError):
        get_verification_hostname("unsupported", "1234567890")
